!      include "mpif.h"
      use mpi
      include "dmemf.h"

      integer   :: i, ierr, size, rank, dummy, dmem_comm
      integer   :: status(MPI_STATUS_SIZE)
!      integer*8 :: memsize, dh0(3), dh1(3), dh2(3), dh4(3)
!      integer*8 :: put_size, offset, len
      integer(kind=MPI_ADDRESS_KIND) :: memsize, dh0(3), dh1(3), dh2(3)
      integer(kind=MPI_ADDRESS_KIND) :: dh4(3)
      integer(kind=MPI_ADDRESS_KIND) :: put_size, offset, len
      integer   :: provided
      integer*1 :: pkt0(1000), pkt1(2000), pkt2(3000)
      integer*1 :: part_buf(20), update_buf(20)
      double precision ask_local, ask_where, transfer, memcpy
      integer   :: num_irecvs

      call MPI_Init_thread(MPI_THREAD_MULTIPLE, provided, ierr)
      if ( ierr /= MPI_SUCCESS ) then
          write(6,*) 'MPI_Init failed with rc=', ierr
      endif

      if (provided /= MPI_THREAD_MULTIPLE) then
         write(6,*) 'DMEM needs MPI_THREAD_MULTIPLE'
         call flush(6)
         call MPI_Abort(MPI_COMM_WOLRD, 444, ierr)
      endif

      call MPI_Comm_size(MPI_COMM_WORLD, size, ierr)
      call MPI_Comm_rank(MPI_COMM_WORLD, rank, ierr)

      if (size /= 3) then
         write(6,*) 'This program needs exactly 3 MPI ranks'
         call flush(6)
         stop
      endif

      call MPI_Comm_dup(MPI_COMM_WORLD, dmem_comm, ierr)

      num_irecvs = 5

      if     (rank == 0) then
         memsize = 0
         call DMEM_Init(dmem_comm, memsize, num_irecvs, ierr)
         call DMEM_Set_errors_return(1, ierr)
         do i = 1, 1000, 1
            pkt0(i) = 48
         enddo
         put_size = 1000
         call DMEM_Put(pkt0, put_size, dh0, ierr)
         call MPI_Bcast(dh0, 3, MPI_AINT, 0, MPI_COMM_WORLD, ierr)
         call MPI_Bcast(dh1, 3, MPI_AINT, 1, MPI_COMM_WORLD, ierr)
         call MPI_Bcast(dh2, 3, MPI_AINT, 2, MPI_COMM_WORLD, ierr)

         call MPI_Recv(dummy, 1, MPI_INT, 2, 47, MPI_COMM_WORLD,          &
     &                 status, ierr)
         offset = 10
         len    = 20
         write(6,*) 'Calling Get_part, handle is'
         call DMEM_Print_handle(dh1, ierr)
         call DMEM_Get_part(dh1, offset, len, part_buf, ierr)
         do i = 1, 20, 1
            if (part_buf(i) /= 52) then
               write(6,*) 'Bad match in Get_part ', part_buf(i),          &
     &                    ' should be 4'
               call flush(6)
               go to 100
            endif
         enddo
 100     continue

      elseif (rank == 1) then
         memsize = 20000
         call DMEM_Init(dmem_comm, memsize, num_irecvs, ierr)
         call DMEM_Set_errors_return(1, ierr)
         do i = 1, 2000, 1
            pkt1(i) = 49
         enddo
         put_size = 2000
         call DMEM_Put(pkt2, put_size, dh1, ierr)
         call MPI_Bcast(dh0, 3, MPI_AINT, 0, MPI_COMM_WORLD, ierr)
         call MPI_Bcast(dh1, 3, MPI_AINT, 1, MPI_COMM_WORLD, ierr)
         call MPI_Bcast(dh2, 3, MPI_AINT, 2, MPI_COMM_WORLD, ierr)

         call DMEM_Get(dh0, pkt0, ierr)
         do i = 1, 1000, 1
            if (pkt0(i) /= 48) then
               write(6,*) 'bad match on pkt0(i), ', pkt0(i),              &
     &                    'should be 0'
               call flush(6)
               go to 200
            endif
         enddo
 200     continue
         call DMEM_Copy(dh2, dh4, ierr)
         call DMEM_Free(dh2, ierr)
         call DMEM_Free(dh4, ierr)
!        test abort vs return code on bad get
         write(6,*) 'Deliberately calling DMEM_Get on freed handle'
         call DMEM_Get(dh4, pkt0, ierr)
         if (ierr == -1) then
            write(6,*) 'Caught deliberate bad address in Get'
         endif
      elseif (rank == 2) then
         memsize = 30000
         call DMEM_Init(dmem_comm, memsize, num_irecvs, ierr)
         call DMEM_Set_errors_return(1, ierr)
         do i = 1, 3000, 1
            pkt2(i) = 50
         enddo
         put_size = 3000
         call DMEM_Put(pkt2, put_size, dh2, ierr)
         call MPI_Bcast(dh0, 3, MPI_AINT, 0, MPI_COMM_WORLD, ierr)
         call MPI_Bcast(dh1, 3, MPI_AINT, 1, MPI_COMM_WORLD, ierr)
         call MPI_Bcast(dh2, 3, MPI_AINT, 2, MPI_COMM_WORLD, ierr)

         do i = 1, 20, 1
            update_buf(i) = 52
         enddo
         offset = 10
         len    = 20
         call DMEM_Update(dh1, offset, len, update_buf, ierr)
         call MPI_Send(dummy, 1, MPI_INT, 0, 47, MPI_COMM_WORLD, ierr)
      endif

      call MPI_Barrier(MPI_COMM_WORLD, ierr)
      if (rank == 0) then
         write (6,*) 'memtable at end of user operations'
         call DMEM_Dumpglobmemtab(ierr)
!        write (6,*) 'after dumping memtable at end of user operations'
!        call DMEM_Abort(777);
      endif

      call DMEM_Finalize(ierr)
      call MPI_Finalize(ierr)

      stop
      end
