/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

/* Exercizing MPI-3 shared memory routines */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "mpi.h"

int main(int argc, char *argv[])
{
    int size, rank, shmemsize, shmemrank;
    MPI_Aint sz_other;
    int dispunit;
    int key = 0;
    MPI_Aint memsize;
    MPI_Comm shmemcomm;
    void *baseptr;
    void *rank0_baseptr;
    MPI_Win shwin;

    MPI_Init(0, 0);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    MPI_Comm_split_type(MPI_COMM_WORLD, MPI_COMM_TYPE_SHARED, key, MPI_INFO_NULL, &shmemcomm);

    MPI_Comm_rank(shmemcomm, &shmemrank);
    MPI_Comm_size(shmemcomm, &shmemsize);
    printf("rank %d of %d in shmemcomm is rank %d in MPI_COMM_WORLD\n", shmemrank, shmemsize, rank);

    /* Process 0 on each node allocates the whole shared memory window */
    if (shmemrank == 0)
        memsize = 1024*1024;
    else
        memsize = 0;

    MPI_Win_allocate_shared(memsize, MPI_BYTE, MPI_INFO_NULL, shmemcomm, &baseptr, &shwin);
    printf("rank %d in shmemcomm got baseptr %p from allocate_shared\n", shmemrank, baseptr);

    // Let's try without for the time being: MPI_Win_lock_all(MPI_MODE_NOCHECK, shwin);

    // Get address of beginning of shared memory acquired by proc 0

    MPI_Win_shared_query(shwin, 0, &sz_other, &dispunit, &rank0_baseptr);
    printf("rank %d has shmemcomm 0's baseptr %p\n", shmemrank, rank0_baseptr);

    /* Test whether the memory is really shared */
    if (shmemrank == 0) {
        *(char*)baseptr = (char) 'a';
        printf("rank 0 in shmemcomm putting 'a' in shared memory\n");
    }

    // MPI_Win_sync(shwin);  // this may be nec if also doing RMA operations on shwin

    // Need this to ensure data is written before being read
    MPI_Barrier(MPI_COMM_WORLD);

    // Use shared memory (prior to maybe unlock_all below)

    printf("rank %d in shmemcomm reading %c from shared memory\n",shmemrank, *(char*)rank0_baseptr);

    // try without for the time being: MPI_Win_unlock_all(shwin);

    MPI_Finalize();

    return 0;
}  
