/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

/* Minimal test application program for the dmem library */

#include "mpi.h"
#include "dmem.h"

static MPI_Datatype handle_type;

int main(int argc, char *argv[])
{
    int i, mpi_rank, mpi_size, provided;
    int num_msgs, msg_size, parms[2];
    MPI_Aint memsize;
    char pkt0[1000], pkt1[2000], pkt2[3000], *buf;
    dmem_handle dh0, dh1, dh2;
    MPI_Comm dmem_comm;

    MPI_Init_thread(0, 0, MPI_THREAD_MULTIPLE, &provided);
    if (provided != MPI_THREAD_MULTIPLE) {
        printf("insufficient level of thread safety provided by MPI implementation\n");
        MPI_Abort(MPI_COMM_WORLD,777);
    }
    MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
    MPI_Comm_dup(MPI_COMM_WORLD, &dmem_comm);

    MPI_Type_vector(3, 1, 1, MPI_AINT, &handle_type);
    MPI_Type_commit(&handle_type);

    /* Each process will handle X messages of size Y, where X and Y are
       command-line arguments */
    
    if (mpi_rank == 0) {        /* assume cmd line is available on rank 0 */
        if (argc != 3) {
            printf("usage: %s <number of messages> < messagesize>\n", argv[0]);
            MPI_Abort(MPI_COMM_WORLD, 999);
        } else {
            num_msgs = atoi(argv[1]);
            msg_size = atoi(argv[2]);
            printf("%d packages of size %d\n", num_msgs, msg_size);
            parms[0] = num_msgs;
            parms[1] = msg_size;
        }
    }
    MPI_Bcast(parms, 2, MPI_INT, 0, MPI_COMM_WORLD);
    num_msgs = parms[0];
    msg_size = parms[1];

    /* Initialize dmem */
    if (mpi_rank == 0)
        DMEM_Init(dmem_comm, -1, 20); /* manager contributes no memory */
    else
        DMEM_Init(dmem_comm, (num_msgs * msg_size), 20);

    MPI_Barrier(dmem_comm);
    if (mpi_rank == 0) {
        printf("Global memory table after initialization\n");
        DMEM_Dumpglobmemtab();
    }
    MPI_Barrier(dmem_comm);

    if (mpi_rank != 0) {
        dmem_handle dh;
        int i;
        char * pkt;

        pkt = (char *) malloc(msg_size);
        for (i = 0; i < num_msgs; i++) {
            DMEM_Put(pkt, msg_size, dh); /* just re-use handle */
            /*
            printf("%d: handle from put: ", mpi_rank);
            DMEM_Print_handle(dh);
            printf("\n");
            */
        }
    }

    MPI_Barrier(MPI_COMM_WORLD);

    if (mpi_rank == 0) {
        printf("Memtable at end of user operations:\n");
        DMEM_Dumpglobmemtab();
    }

    DMEM_Finalize();
    MPI_Finalize();
    return 0;
}  
  
